// name: RotMethyl.C
// author: J. Michael Word
// date written: 2/7/98
// purpose: Implementation for RotMethyl

// **************************************************************
// NOTICE: This is free software and the source code is freely
// available. You are free to redistribute or modify under the
// conditions that (1) this notice is not removed or modified
// in any way and (2) any modified versions of the program are
// also available for free.
//               ** Absolutely no Warranty **
// Copyright (C) 1999 J. Michael Word
// **************************************************************

#ifdef OLD_STD_HDRS
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#else
#include <cstdio>
#include <cctype>
#include <cmath>
#endif

#include "RotMethyl.h"
#include "AtomPositions.h"

#define START_ANGLE 180.0
#define ROUGH_STEP   30
#define FINE_STEP     1

RotMethyl::RotMethyl(const Point3d& a, const Point3d& b,
                     const double ang, const PDBrec& heavyAtom)
   : _p1(a), _p2(b), _angle(ang), _heavyAtom(heavyAtom) {
   strcpy(_grpName, ((heavyAtom.elem().atno() == 7) ?
			"NH3+   " : "methyl "));
   validateMemo();
}

void RotMethyl::finalize(int nBondCutoff, bool,
      AtomPositions &xyz, DotSphManager&) {

   if (isComplete()) {

      // pre-build lists of bonded atoms

      const double approxNbondDistLimit = 3.0 + 0.5*nBondCutoff; // just a rule of thumb
      _bnded.resize(_rot.len()+1);
      int i = 0;
      
      Seq<PDBrec> rotAtmList(_heavyAtom, _rot);
      for(Seq<PDBrec> alst = rotAtmList; alst; alst++) {
	 const PDBrec& thisAtom = *alst;
	 _bnded[i++] = bondedList(thisAtom,
	                          xyz.neighbors(thisAtom.loc(),
	                                        thisAtom.covRad(),
	                                        approxNbondDistLimit),
	                          nBondCutoff, rotAtmList);
      }
   }
}

int RotMethyl::makebumpers(MultiDict<LocBlk, BumperPoint>& bblks,
      int rn, float& maxVDWrad) {
   int an = 0;
   const double dtheta = 10.0; // fineness of rotation angle scan
   
   const double scanAngle = 60.0;
   for (Seq<PDBrec> a = _rot; a; a++) {
      for (double theta = -scanAngle; theta < scanAngle; theta += dtheta) {
	 Point3d p((*a).loc().rotate(theta, _p2, _p1));
	 bblks.put(LocBlk(p), BumperPoint(p, rn, an++, (*a).vdwRad()));
	 if ((*a).vdwRad() > maxVDWrad) { maxVDWrad = (*a).vdwRad(); }
      }
   }
   return an;
}

int RotMethyl::numOrientations(SearchStrategy ss) const {
   return (ss==Mover::LOW_RES) ?
      int(120.0/ROUGH_STEP + 0.5) :
      int(2.0*(double(ROUGH_STEP)/double(FINE_STEP)) + 0.5);
}

bool RotMethyl::isDefaultO(int oi, SearchStrategy ss) const {
   if (ss!=Mover::LOW_RES) { oi = orientation(Mover::LOW_RES); }
   return ((oi == 0) && 
           (abs(clampAngle(START_ANGLE - angle())) < 1.0));
}

bool RotMethyl::setOrientation(int oi, float delta, AtomPositions &xyz,
                                                     SearchStrategy ss) {

   const double oldTheta = angle();
   const double    theta = orientationAngle(oi, ss) + delta;

   if (abs(theta-oldTheta) > 0.1) {
      for(Seq<PDBrec> hydlist = _rot; hydlist; hydlist++) {
	 setHydAngle(*hydlist, oldTheta, theta, xyz);
      }
      angle(theta);
   }
   rememberOrientation(oi, ss);
   return TRUE;
}

double RotMethyl::orientationAngle(int oi, SearchStrategy ss) const {
   double delta = 0.0;
   if (ss != Mover::LOW_RES) {
      const int oh = oi;
      oi = orientation(Mover::LOW_RES);
      if (oi == 0) {   // centered on START_ANGLE
	 delta = ((oh+1)>>1) * ((oh&1) ? 1.0 : -1.0) * FINE_STEP;
      }
      else {
	 const int halfRange=int(double(ROUGH_STEP)/double(FINE_STEP) + 0.5);

	 if (oi&1) { // odd  - angles beyond START_ANGLE
	    delta = (oh * FINE_STEP) - halfRange;
	 }
	 else {      // even - angles before START_ANGLE
	    delta = halfRange - (oh * FINE_STEP);
	 }
      }
   }

   const double a = START_ANGLE + delta +
	    ((oi+1)>>1) * ((oi&1) ? 1.0 : -1.0) * ROUGH_STEP;
   return clampAngle(a);
}

Stringclass RotMethyl::describeOrientation() const {
   char descrbuf[25];
   ::sprintf(descrbuf,"%s%4.0f",_grpName, _angle);
   return Stringclass(descrbuf);
}

double RotMethyl::determineScore(AtomPositions &xyz,
   DotSphManager& dotBucket, int nBondCutoff, float probeRadius,
   float pmag, double& penalty, float &bumpScore, float &hbScore, bool &hasBadBump) {

   bumpScore  = 0.0;
   hbScore    = 0.0;
   hasBadBump = FALSE;

   double bestScore = scoreThisAngle(xyz, dotBucket,
                         nBondCutoff, probeRadius,
                         bumpScore, hbScore,
			 hasBadBump);

   penalty = orientationPenalty(pmag);
   return bestScore;
}

double RotMethyl::scoreThisAngle(AtomPositions &xyz,
   DotSphManager& dotBucket, int, float probeRadius,
   float &bumpScore, float &hbScore, bool &hasBadBump) {
   const double maxVDWrad = ElementInfo::StdElemTbl().maxExplicitRadius();

   bumpScore  = 0.0;
   hbScore    = 0.0;
   hasBadBump = FALSE;

   double scoreThisO = 0.0;
   int i = 0;
   for(Seq<PDBrec> alst(_heavyAtom, _rot); alst; alst++) {
      float bumpSubScore = 0.0;
      float hbSubScore   = 0.0;
      bool subBadBump    = FALSE;
      const PDBrec& thisAtom = *alst;
      Seq<PDBrec> nearby = xyz.neighbors(thisAtom.loc(), 0.001,
				 thisAtom.vdwRad() + probeRadius + maxVDWrad);

      double val = xyz.atomScore(thisAtom, thisAtom.loc(), nearby,
		_bnded[i], dotBucket.fetch(thisAtom.vdwRad()), probeRadius, FALSE,
		bumpSubScore, hbSubScore, subBadBump);

      bumpScore  += bumpSubScore;
      hbScore    += hbSubScore;
      if (subBadBump) { hasBadBump = TRUE; }

#ifdef DEBUGSUBSCORE
   cerr << "\t:" << PDBrecNAMEout(thisAtom)
      <<":"<< describeOrientation()
      << ": bump=" << bumpSubScore
      << ", HB=" << hbSubScore
      << ((subBadBump==TRUE)?", BADBUMP":"")
      << "\t"
//      << thisAtom.loc()
      << endl;
#endif
      scoreThisO += val;
      i++;
   }

   initializeScoreIfNotSet(scoreThisO, hasBadBump);
   return scoreThisO;
}

#ifdef ROTPENALTY
double RotMethyl::orientationPenalty(float pmag) const {
   const double freq = 1.5*(PI/180.0);
   // relatively small penalty
   return -pmag * 0.1 * abs(sin( freq*(angle()-START_ANGLE) ));
}
#else
double RotMethyl::orientationPenalty(float) const {
   return 0.0;
}
#endif

void RotMethyl::setHydAngle(double newAng, AtomPositions &xyz) {
   const double oldAng = angle();
   for(Seq<PDBrec> hydlist = _rot; hydlist; hydlist++) {
      setHydAngle(*hydlist, oldAng, newAng, xyz);
//cerr  << "RotMethyl::setHydAngle[" << (*hydlist).atomname() << "] "
//      << oldAng << " deg --> "
//      << newAng << " deg" << endl;
   }
   angle(newAng);
}

void RotMethyl::setHydAngle(PDBrec theAtom, double oldAng, double newAng,
                        AtomPositions &xyz) {
   if (abs(newAng-oldAng) > 0.0001) {
      Point3d lastloc = theAtom.loc();
      theAtom.loc( lastloc.rotate(newAng-oldAng, _p2, _p1) );
      xyz.reposition(lastloc, theAtom);
   }
}
